------------------------------------------------------
-- FeedOMatic.lua
-- Version 1700.4
------------------------------------------------------

-- TODO: if you don't auto-feed, have an option to make the need to feed more noticeable (e.g., pulsing halo around the pet-happiness icon).
-- TODO: don't feed pet while he's under a DOT (poison etc) -- the damage ticks interrupt feed effect and cause unhappiness

BINDING_HEADER_FEEDOMATIC = "Fizzwidget Feed-O-Matic";
BINDING_NAME_FEEDOMATIC = "Feed Pet";

-- constants
OUT_OF_FOOD_TIMEOUT = 60 * 5; -- don't try to autofeed more than once every five minutes if out of (usable) food
MAX_QUALITY = 35 * 60 + 1; -- We store a notion of a food's "quality": its best happiness-per-tick multiplied by the pet's level as of when that tick occurred. We use "best" because a pet that's closer to "sated" (maximum happiness) will receive less happiness per tick than he would from the same food if he were hungrier. (So, a food that gives 35 happiness per tick to a level 60 pet is "better" than a food that's worth 35 happiness per tick to a level 30 pet.) Foods whose quality hasn't been observed yet are given this value when sorting, so we can prioritize the discovery of new foods' quality ratings.
MAX_KEEPOPEN_SLOTS = 150;

-- Configuration
FOM_Config_Default = {
	Enabled = false;
	Alert = "emote";
	Level = "content";
	KeepOpenSlots = 8;
	AvoidUsefulFood = true;
	AvoidQuestFood = true;
	AvoidBonusFood = true;
	Fallback = true;
	SaveForCookingLevel = 1;
	PreferHigherQuality = true;
};
FOM_Config = FOM_Config_Default;

-- Foods Lists (for reference only; loaded from FOM_Foods.lua)
-- FOM_Foods = { };
-- FOM_Foods["meat"] = { };
-- FOM_Foods["fish"] = { };
-- FOM_Foods["bread"] = { };
-- FOM_Foods["cheese"] = { };
-- FOM_Foods["fruit"] = { };
-- FOM_Foods["fungus"] = { };
-- FOM_Foods["bonus"] = { };

-- FOM_Cooking = { };
-- Has the following internal structure:
--		REALM_PLAYER = {
--			FOODNAME = SKILL_DIFFICULTY,
--		}

-- FOM_QuestFood = { };
-- Has the following internal structure:
--		REALM_PLAYER = {
--			FOODNAME = QUANTITY_REQUIRED,
--		}

-- FOM_FoodQuality = { };
-- Has the following internal structure:
--		REALM_PLAYER = {
--			PETNAME = {
--				FOODNAME = HAPPINESS,
--			}
--		}

-- Variables
FOM_State = { };
FOM_State.InCombat = false;
FOM_State.IsAFK = false;
FOM_State.ShouldFeed = false;
FOM_State.OutOfFoodTime = OUT_OF_FOOD_TIMEOUT;

local LastFood = nil;
local RealmPlayer = nil;
local LastPetName = nil;

-- Anti-freeze code borrowed from ReagentInfo (in turn, from Quest-I-On):
-- keeps WoW from locking up if we try to scan the tradeskill window too fast.
FOM_TradeSkillLock = { };
FOM_TradeSkillLock.Locked = false;
FOM_TradeSkillLock.EventTimer = 0;
FOM_TradeSkillLock.EventCooldown = 0;
FOM_TradeSkillLock.EventCooldownTime = 1;

-- ... and repurposed to keep us from feeding twice on updates.
FOM_FeedLock = { };
FOM_FeedLock.Locked = false;
FOM_FeedLock.EventTimer = 0;
FOM_FeedLock.EventCooldown = 0;
FOM_FeedLock.EventCooldownTime = 5;

-- State variable used to track required quantities of quest food when it's in more than one stack
FOM_Quantity = { };

function FOM_OnLoad()

	-- Register for Events
	this:RegisterEvent("VARIABLES_LOADED");

	-- Register Slash Commands
	SLASH_FEEDOMATIC1 = "/feedomatic";
	SLASH_FEEDOMATIC2 = "/fom";
	SLASH_FEEDOMATIC3 = "/feed";
	SLASH_FEEDOMATIC4 = "/petfeed"; -- Rauen's PetFeed compatibility
	SLASH_FEEDOMATIC5 = "/pf";
	SlashCmdList["FEEDOMATIC"] = function(msg)
		FOM_ChatCommandHandler(msg);
	end
	
	-- hook functions so we can manage per-pet saved food quality data
	FOM_Original_PetRename = PetRename;
	PetRename = FOM_PetRename;
	FOM_Original_PetAbandon = PetAbandon;
	PetAbandon = FOM_PetAbandon;
	
	--GFWUtils.Debug = true;

	GFWUtils.Print("Fizzwidget Feed-O-Matic initialized!");

end

function FOM_CheckSetup()

	_, realClass = UnitClass("player");
	if (realClass ~= "HUNTER") then return; end

	if (RealmPlayer == nil) then
		RealmPlayer = GetCVar("realmName") .. "." .. UnitName("player");
	end
	
	if (FOM_FoodQuality == nil) then
		FOM_FoodQuality = { };
	end	
	if (FOM_FoodQuality[RealmPlayer] == nil) then
		FOM_FoodQuality[RealmPlayer] = { };
	end
	LastPetName = UnitName("pet");
	if (LastPetName and LastPetName ~= "" and LastPetName ~= UNKNOWNOBJECT) then
		if (FOM_FoodQuality[RealmPlayer][LastPetName] == nil) then
			FOM_FoodQuality[RealmPlayer][LastPetName] = { };
		end
	end
	
end

function FOM_OnEvent(event, arg1)

	-- Save Variables
	if ( event == "VARIABLES_LOADED" ) then
		
		FOM_LoadOldData();
		
		_, realClass = UnitClass("player");
		if (realClass == "HUNTER") then
			-- monitor status for whether we're able to feed
			this:RegisterEvent("PET_ATTACK_START");
			this:RegisterEvent("PET_ATTACK_STOP");
			this:RegisterEvent("CHAT_MSG_SYSTEM");
						
			-- check your pet roster when at the stables so we don't bloat SavedVariables
			this:RegisterEvent("PET_STABLE_SHOW");
			this:RegisterEvent("PET_STABLE_UPDATE");

			-- track whether foods are useful for Cooking 
			this:RegisterEvent("TRADE_SKILL_SHOW");
			this:RegisterEvent("TRADE_SKILL_UPDATE");

			-- figure out what happens when we try to feed pet (gain happiness, didn't like, etc)
			this:RegisterEvent("CHAT_MSG_SPELL_TRADESKILLS");
			this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS");
			this:RegisterEvent("UI_ERROR_MESSAGE");
			this:RegisterEvent("UI_INFO_MESSAGE");
			
			-- Events for trying to catch when the pet needs feeding
			this:RegisterEvent("PET_BAR_SHOWGRID");
			this:RegisterEvent("PET_BAR_UPDATE");
			this:RegisterEvent("PET_UI_UPDATE");
			this:RegisterEvent("UNIT_HAPPINESS");

		end
		return;
		
	elseif ( event == "CHAT_MSG_SYSTEM") then
	
		local msg = arg1;
		
		if ( string.sub(msg, 1, 16) == "You are now AFK:" ) then
			FOM_State.IsAFK = true;
			return;
		end
		
		if ( string.sub(msg, 1, 22) == "You are no longer AFK." ) then
			FOM_State.IsAFK = false;
		end
		
	elseif ( event == "PET_ATTACK_START" ) then
	
		-- Set Flag
		FOM_State.InCombat = true;
		return;
		
	elseif ( event == "PET_ATTACK_STOP" ) then
	
		-- Remove Flag
		FOM_State.InCombat = false;
		
	elseif ( event == "CHAT_MSG_SPELL_TRADESKILLS" ) then
	
		_, _, foodName = string.find(arg1, "Your pet begins eating the ([%w ]+).");
		if (foodName and foodName ~= "") then
			GFWUtils.DebugLog("Manually fed "..foodName);
			LastFood = foodName;
		end
		return;

	elseif ( event == "CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS" ) then
		
		if (arg1 and FOM_HasFeedEffect()) then
			_, _, happiness = string.find(arg1, "gains (%d+) Happiness");
			if (happiness == nil or happiness == "") then return; end
			happiness = tonumber(happiness); --convert to number 
			
			FOM_CheckSetup();
			
			if (LastFood ~= nil and happiness > 0) then
				local knownQuality = FOM_FoodQuality[RealmPlayer][LastPetName][LastFood];
				local quality = happiness * UnitLevel("pet");
				FOM_FoodQuality[RealmPlayer][LastPetName][LastFood] = math.max((knownQuality or 0), quality);
				LastFood = nil;
			end
		end
		return;
	
	elseif ( event == "UI_ERROR_MESSAGE" ) then

		if (arg1 and string.find(arg1, "level is not high enough for your pet")) then
			
			FOM_CheckSetup();

			if not (LastFood == nil and FOM_FoodQuality[RealmPlayer][LastPetName][LastFood] ~= 0) then
				FOM_FoodQuality[RealmPlayer][LastPetName][LastFood] = 0;
				LastFood = nil;
				if ( FOM_Config.Alert == "chat") then
					GFWUtils.Print(UnitName("pet").." doesn't like that food and will try something else...");
				elseif ( FOM_Config.Alert == "emote") then
					SendChatMessage("looks for something else to feed ".. UnitName("pet") .."...", "EMOTE");
				end
				FOM_State.ShouldFeed = true;
				--FOM_Feed(nil);
				return;
			end
		
		elseif (arg1 and string.find(arg1, "doesn't like that food")) then
			
			FOM_CheckSetup();

			if not (LastFood == nil) then

				if ( FOM_Config.Alert == "chat") then
					GFWUtils.Print(UnitName("pet").." doesn't like that food and will try something else...");
				elseif ( FOM_Config.Alert == "emote") then
					SendChatMessage("looks for something else to feed ".. UnitName("pet") .."...", "EMOTE");
				end
				-- remove from quality tracking
				FOM_FoodQuality[RealmPlayer][LastPetName][LastFood] = nil;
				
				-- remove from diet
				local dietList = {GetPetFoodTypes()};
				for _, diet in dietList do 
					if ( FOM_RemoveFood(string.lower(diet), LastFood) ) then
						local capDiet = string.upper(string.sub(diet, 1, 1)) .. string.sub(diet, 2); -- print a nicely capitalized version
						GFWUtils.Print("Removed "..GFWUtils.Hilite(LastFood).." from "..GFWUtils.Hilite(capDiet).." list.");
					end
				end
				LastFood = nil;
				
				FOM_State.ShouldFeed = true;
				--FOM_Feed(nil);
				return;
			end
		end
		return;
		
	elseif (event == "TRADE_SKILL_SHOW" or event == "TRADE_SKILL_UPDATE" and GetTradeSkillLine() == "Cooking") then

		if (FOM_Config.SaveForCookingLevel >= 0 and FOM_Config.SaveForCookingLevel <= 3) then
			-- Update Cooking reagents list so we can avoid consuming food we could skillup from.
			if (RealmPlayer == nil) then
				RealmPlayer = GetCVar("realmName") .. "." .. UnitName("player");
			end
			if (FOM_Cooking == nil) then
				FOM_Cooking = { };
			end
			if (FOM_Cooking[RealmPlayer] == nil) then
				FOM_Cooking[RealmPlayer] = { };
			end
			if (FOM_Cooking[RealmPlayer] ~= nil and TradeSkillFrame:IsVisible() and not FOM_TradeSkillLock.Locked) then
				-- This prevents further update events from being handled if we're already processing one.
				-- This is done to prevent the game from freezing under certain conditions.
				FOM_TradeSkillLock.Locked = true;

				FOM_CheckSetup();
				for i=1, GetNumTradeSkills() do
					local itemName, type, _, _ = GetTradeSkillInfo(i);
					if (type ~= "header") then
						for j=1, GetTradeSkillNumReagents(i) do
							local reagentName, _, _, _ = GetTradeSkillReagentInfo(i, j);

							if (reagentName ~= nil and FOM_IsKnownFood(reagentName)) then
								if (FOM_Cooking[RealmPlayer][reagentName] == nil) then
									FOM_Cooking[RealmPlayer][reagentName] = FOM_DifficultyToNum(type);
								else
									FOM_Cooking[RealmPlayer][reagentName] = max(FOM_Cooking[RealmPlayer][reagentName], FOM_DifficultyToNum(type));
								end
							end
						end
					end
				end
			end
		end
		return;
		
	elseif (event == "PET_STABLE_SHOW" or event == "PET_STABLE_UPDATE") then
	
		-- clean up the FOM_FoodQuality sub-tables in case we missed you abandoning a pet
		FOM_CheckSetup();
		local stabledPetNames = {};
		for petIndex = 0, 2 do
			local _, petName, _, _, _ = GetStablePetInfo(petIndex);
			if (petName) then
				table.insert(stabledPetNames, petName);
			end
		end
		local orphanedPetNames = {};
		for savedPetName in FOM_FoodQuality[RealmPlayer] do
			if (stabledPetNames == nil) then
				GFWUtils.DebugLog("stabledPetNames == nil");
			end
			if (stabledPetNames ~= nil and GFWTable.IndexOf(stabledPetNames, savedPetName) == 0) then
				table.insert(orphanedPetNames, savedPetName);
			end
		end
		for orphanedPet in orphanedPetNames do
			FOM_FoodQuality[RealmPlayer][orphanedPet] = {};
		end
		return;

	end

	-- Check Happiness (and feed if necessary)
	if ( FOM_Config.Enabled ) then
		FOM_CheckHappiness();
	end
	
end

-- Update our list of quest objectives so we can avoid consuming food we want to accumulate for a quest.
function FOM_ScanQuests()
	for questNum=1, GetNumQuestLogEntries() do
		local QText, level, questTag, isHeader, isCollapsed, isComplete  = GetQuestLogTitle(questNum);
		if (not isHeader) then
			for objectiveNum=1, GetNumQuestLeaderBoards(questNum) do
				local text, type, finished = GetQuestLogLeaderBoard(objectiveNum, questNum);
				if (text ~= nil and strlen(text) > 0) then
					local _, _, objectiveName, numCurrent, numRequired = string.find(text, "(.*): (%d+)/(%d+)");
					if (FOM_IsKnownFood(objectiveName)) then
						if (FOM_QuestFood == nil) then
							FOM_QuestFood = { };
						end
						if (FOM_QuestFood[RealmPlayer] == nil) then
							FOM_QuestFood[RealmPlayer] = { };
						end

						if (FOM_QuestFood[RealmPlayer][objectiveName] == nil) then
							FOM_QuestFood[RealmPlayer][objectiveName] = tonumber(numRequired);
						else
							FOM_QuestFood[RealmPlayer][objectiveName] = max(FOM_QuestFood[RealmPlayer][objectiveName], tonumber(numRequired));
						end
					end
				end
			end
		end
	end
end

function FOM_DifficultyToNum(level)
	if (level == "optimal" or level == "orange") then
		return 3;
	elseif (level == "medium" or level == "yellow") then
		return 2;
	elseif (level == "easy" or level == "green") then
		return 1;
	elseif (level == "trivial" or level == "gray" or level == "grey") then
		return 1;
	else -- bad input
		return nil;
	end
end

function FOM_OnUpdate(elapsed)
	
	_, realClass = UnitClass("player");
	if (realClass ~= "HUNTER") then return; end

	-- If it's been more than a second since our last tradeskill update,
	-- we can allow the event to process again.
	FOM_TradeSkillLock.EventTimer = FOM_TradeSkillLock.EventTimer + elapsed;
	if (FOM_TradeSkillLock.Locked) then
		FOM_TradeSkillLock.EventCooldown = FOM_TradeSkillLock.EventCooldown + elapsed;
		if (FOM_TradeSkillLock.EventCooldown > FOM_TradeSkillLock.EventCooldownTime) then

			FOM_TradeSkillLock.EventCooldown = 0;
			FOM_TradeSkillLock.Locked = false;
		end
	end
	
	-- using a similar mechanism to keep us from feeding twice in one update.
	FOM_FeedLock.EventTimer = FOM_FeedLock.EventTimer + elapsed;
	if (FOM_FeedLock.Locked) then
		FOM_FeedLock.EventCooldown = FOM_FeedLock.EventCooldown + elapsed;
		if (FOM_FeedLock.EventCooldown > FOM_FeedLock.EventCooldownTime) then

			FOM_FeedLock.EventCooldown = 0;
			FOM_FeedLock.Locked = false;
		end
	end
	
	--GFWUtils.Debug = true;

	if (FOM_State.ShouldFeed == true) then
		--DevTools_Dump(FOM_FeedLock);
	end
	if (FOM_State.ShouldFeed == true and FOM_FeedLock.Locked == false and FOM_CanFeed() ) then
		FOM_FeedLock.Locked = true;
		FOM_State.ShouldFeed = false;
		GFWUtils.DebugLog("Feeding on update");
		if (GetTime() - FOM_State.OutOfFoodTime > OUT_OF_FOOD_TIMEOUT) then
			FOM_Feed(nil);
		end
	end
end

function FOM_CanFeed()
	if ( not UnitExists("pet") ) then
		GFWUtils.DebugLog("Can't feed; pet doesn't exist.");
		return false;
	end
	if ( UnitHealth("pet") <= 0 ) then
		GFWUtils.DebugLog("Can't feed; pet is dead.");
		return false;
	end
	if ( UnitHealth("player") <= 0 ) then
		GFWUtils.DebugLog("Can't feed; I'm dead.");
		return false;
	end
	if ( CastingBarFrameStatusBar:IsVisible() ) then
		GFWUtils.DebugLog("Can't feed; casting a spell / tradeksill.");
		return false;
	end
	if ( UnitOnTaxi("player") ) then
		GFWUtils.DebugLog("Can't feed; flying.");
		return false;
	end
	if ( FOM_State.InCombat ) or ( PlayerFrame.inCombat ) then
		GFWUtils.DebugLog("Can't feed; in combat.");
		return false;
	end
	
	local buff, buffIndex;
	local dontFeedBuffTextures = { 
		"Interface\\Icons\\Ability_Ambush",				-- NE Shadowmeld (maybe not unique buff icon?)
		"Interface\\Icons\\Ability_Rogue_FeignDeath",	-- Feign Death
		"Interface\\Icons\\INV_Drink_07",				-- drinking
		"Interface\\Icons\\INV_Misc_Fork&Knife",		-- eating
	};
	local mountTextureSubStrings = { 
		"Ability_Mount",
		"INV_Misc_Foot_Kodo",
	};
	local mountNames = { "kodo", "wolf", "raptor", "steed", "horse", "ram", "saber", "mechanostrider" };
	for buffIndex=0, 15 do
		buff = GetPlayerBuffTexture(buffIndex);
		if ( buff ~= nil) then
			for _, buffTexture in dontFeedBuffTextures do
				if ( buff == buffTexture ) then
					GFWUtils.DebugLog("Can't feed; currently, eating, drinking, or feigning death.");
					return false;
				end
			end
			if ( UnitLevel("player") >= 40 ) then
				for _, buffTexture in mountTextureSubStrings do
					if ( string.find(buff, buffTexture) ) then
						FOMTooltip:SetUnitBuff("player", buffIndex+1);
						local msg = FOMTooltipTextLeft1:GetText();
						if (msg ~= nil) then
							msg = string.lower(msg);
							for _, mountName in mountNames do
								if (string.find(msg, mountName)) then
									GFWUtils.DebugLog("Can't feed; mounted.");
									return false;
								end
							end
						end
					end
				end
			end
		end
	end
	return true;
end

function FOM_ChatCommandHandler(msg)

	-- Check for Pet (we don't really need one for most of our chat commands, but we conveniently use its name.)
	if ( UnitExists("pet") ) then
		petName = UnitName("pet");
	else
		petName = "Your pet";
	end
	
	-- Print Help
	if ( msg == "help" ) or ( msg == "" ) then
		GFWUtils.Print("Fizzwidget Feed-O-Matic:");
		GFWUtils.Print("/feedomatic /fom <command>");
		GFWUtils.Print("- "..GFWUtils.Hilite("help").." - Print this helplist.");
		GFWUtils.Print("- "..GFWUtils.Hilite("status").." - Check current settings.");
		GFWUtils.Print("- "..GFWUtils.Hilite("reset").." - Reset to default settings.");
		GFWUtils.Print("- "..GFWUtils.Hilite("alert chat").." | "..GFWUtils.Hilite("emote").." | "..GFWUtils.Hilite("off").." - Alert via chat window or emote channel when feeding.");
		GFWUtils.Print("- "..GFWUtils.Hilite("on").." | "..GFWUtils.Hilite("off").." - Turn automatic feeding on or off.");
		GFWUtils.Print("- "..GFWUtils.Hilite("level content").." | "..GFWUtils.Hilite("happy").." - Set happiness level for automatic feeding.");
		GFWUtils.Print("- "..GFWUtils.Hilite("saveforcook orange").." | "..GFWUtils.Hilite("yellow").." | "..GFWUtils.Hilite("green").." | "..GFWUtils.Hilite("gray").." | "..GFWUtils.Hilite("off").." - Avoid foods used in cooking recipes (based on their difficulty).");
		GFWUtils.Print("- "..GFWUtils.Hilite("savequest on").." | "..GFWUtils.Hilite("off").." - Avoid foods you need to collect for a quest.");
		GFWUtils.Print("- "..GFWUtils.Hilite("savebonus on").." | "..GFWUtils.Hilite("off").." - Avoid foods which have bonus effects.");
		GFWUtils.Print("- "..GFWUtils.Hilite("keepopen <number>").." - Set when to prefer smaller stacks of food vs. more nutritious food (see readme).");
		GFWUtils.Print("- "..GFWUtils.Hilite("feed").." - Feed your pet (automatically finds an appropriate food).");
		GFWUtils.Print("- "..GFWUtils.Hilite("feed <name>").." - Feed your pet a specific food.");
		GFWUtils.Print("- "..GFWUtils.Hilite("add <diet> <name>").." - Add food to list.");
		GFWUtils.Print("- "..GFWUtils.Hilite("remove <diet> <name>").." - Remove food from list.");
		GFWUtils.Print("- "..GFWUtils.Hilite("show <diet>").." - Show food list.");
		return;
	end
	
	-- Check Status
	if ( msg == "status" ) then
		if ( FOM_Config.Enabled ) then
			GFWUtils.Print("Automatic feeding is enabled.");
			GFWUtils.Print(petName.." will eat food when less than "..GFWUtils.Hilite(FOM_Config.Level)..".");
		else
			GFWUtils.Print("Automatic feeding is disabled.");
		end
		if (FOM_Config.KeepOpenSlots < MAX_KEEPOPEN_SLOTS) then
			if (FOM_Config.PreferHigherQuality) then
				GFWUtils.Print("Feed-O-Matic will prefer to use higher quality foods first.");
			else
				GFWUtils.Print("Feed-O-Matic will prefer to use lower quality foods first.");
			end
			if (FOM_Config.KeepOpenSlots == 0) then
				GFWUtils.Print("Feed-O-Matic will look first at food quality when determining what to feed to your pet.");
			else
				GFWUtils.Print("If fewer than "..GFWUtils.Hilite(FOM_Config.KeepOpenSlots).." spaces are open in your inventory, Feed-O-Matic will prefer smaller stacks of food regardless of quality.");
			end
		else
			GFWUtils.Print("Feed-O-Matic will always prefer smaller stacks of food regardless of quality.");
		end
		
		if (FOM_Config.Alert == "emote") then
			GFWUtils.Print("You will automatically emote when feeding "..petName..".");
		elseif (FOM_Config.Alert == "chat") then
			GFWUtils.Print("Feed-O-Matic will notify you in chat when feeding "..petName..".");
		else
			GFWUtils.Print("There will be no alert when feeding "..petName..".");
		end
				
		if (FOM_Config.SaveForCookingLevel >= 0 and FOM_Config.SaveForCookingLevel <= 3) then
			if (FOM_Config.SaveForCookingLevel == 3) then
				level = "orange";
			elseif (FOM_Config.SaveForCookingLevel == 2) then
				level = "yellow";
			elseif (FOM_Config.SaveForCookingLevel == 1) then
				level = "green";
			elseif (FOM_Config.SaveForCookingLevel == 0) then
				level = "gray";
			end
			GFWUtils.Print("Feed-O-Matic will avoid foods used in "..GFWUtils.Hilite(level).." or higher Cooking recipes.");
		else
			GFWUtils.Print("Feed-O-Matic will choose foods without regard to whether they're used in Cooking.");
		end
		
		if (FOM_Config.AvoidQuestFood) then
			GFWUtils.Print("Feed-O-Matic will avoid foods you need to collect for quests.");
		else
			GFWUtils.Print("Feed-O-Matic will choose foods without regard to whether they're needed for quests.");
		end
		if (FOM_Config.AvoidBonusFood) then
			GFWUtils.Print("Feed-O-Matic will avoid foods that have an additional bonus effect when eaten by a player.");
		else
			GFWUtils.Print("Feed-O-Matic will choose foods without regard to whether they have bonus effects.");
		end
		if (FOM_Config.Fallback) then
			GFWUtils.Print("Feed-O-Matic will fall back to food it would otherwise avoid if no other food is available.");
		else
			GFWUtils.Print("Feed-O-Matic will not feed your pet if the only foods available are foods you'd prefer to avoid feeding.");
		end
		return;
	end

	-- Reset Variables
	if ( msg == "reset" ) then
		FOM_Config = FOM_Config_Default;
		GFWUtils.Print("Feed-O-Matic configuration reset.");
		FOM_ChatCommandHandler("status");
		return;
	end
	
	-- Toggle Alert
	if ( string.sub(msg, 1, 5) == "alert" ) then
		local _, _, option = string.find(msg, "alert (.+)");
		
		if (option == "emote") then
			FOM_Config.Alert = "emote";
			GFWUtils.Print("You will automatically emote when feeding "..petName..".");
		elseif (option == "chat") then
			FOM_Config.Alert = "chat";
			GFWUtils.Print("Feed-O-Matic will notify you in chat when feeding "..petName..".");
		elseif (option == "off") then
			FOM_Config.Alert = nil;
			GFWUtils.Print("There will be no alert when feeding "..petName..".");
		else
			GFWUtils.Print("Usage: "..GFWUtils.Hilite("/feedomatic alert chat").." | "..GFWUtils.Hilite("emote").." | "..GFWUtils.Hilite("off"));
		end
		return;
	end
	
	-- Turn automatic feeding On
	if ( msg == "on" ) then
		FOM_Config.Enabled = true;
		GFWUtils.Print("Automatic feeding is enabled.");
		FOM_CheckHappiness();
		return;
	end
	
	-- Turn automatic feeding Off
	if ( msg == "off" ) then
		FOM_Config.Enabled = false;
		GFWUtils.Print("Automatic feeding is disabled.");
		return;
	end
	
	-- Set Happiness Level
	if ( string.sub(msg, 1, 5) == "level" ) then
		if ( string.find( msg, "content" ) ) then
			FOM_Config.Level = "content" ;
		elseif ( string.find( msg, "happy" ) ) then
			FOM_Config.Level = "happy" ;
		elseif ( string.find( msg, "debug" ) ) then
			FOM_Config.Level = "debug" ;
		else
			return;
		end
		GFWUtils.Print(petName.." will now eat when less than "..GFWUtils.Hilite(FOM_Config.Level)..".");
		FOM_CheckHappiness();
		return;
	end
	
	local _, _, cmd, option = string.find(msg, "(%w+) (%w+)");

	-- Set Cooking recipe level
	if ( cmd == "saveforcook" ) then
		local level = option;
		if (level ~= nil) then
			local levelNum = FOM_DifficultyToNum(level);
			if (levelNum ~= nil) then 
				FOM_Config.SaveForCookingLevel = levelNum;
				FOM_Config.AvoidUsefulFood = true;
				GFWUtils.Print("Feed-O-Matic will avoid foods used in "..GFWUtils.Hilite(level).." or higher Cooking recipes. You'll need to open your Cooking window for Feed-O-Matic to cache information about what recipes you know.");
				return;
			elseif (level == "off") then
				FOM_Config.SaveForCookingLevel = 4; 
				if (not FOM_Config.AvoidQuestFood and not FOM_Config.Avoid9) then
					FOM_Config.AvoidUsefulFood = false;
				end
				GFWUtils.Print("Feed-O-Matic will choose foods without regard to whether they're used in Cooking.");
				return;
			end
		end
		GFWUtils.Print("Usage: "..GFWUtils.Hilite("/feedomatic saveforcook orange").." | "..GFWUtils.Hilite("yellow").." | "..GFWUtils.Hilite("green").." | "..GFWUtils.Hilite("gray").." | "..GFWUtils.Hilite("off"));
		return;
	end
	
	-- Set avoiding food with bonuses
	if ( cmd == "savequest" ) then
		if (option == "on") then
			FOM_Config.AvoidQuestFood = true;
			FOM_Config.AvoidUsefulFood = true;
			FOM_ScanQuests();
			GFWUtils.Print("Feed-O-Matic will avoid foods you need to collect for quests.");
		elseif (option == "off") then
			FOM_Config.AvoidQuestFood = true;
			if not (FOM_Config.SaveForCookingLevel >= 0 and FOM_Config.SaveForCookingLevel <= 3 and not FOM_Config.AvoidBonusFood) then
				FOM_Config.AvoidUsefulFood = false;
			end
			GFWUtils.Print("Feed-O-Matic will choose foods without regard to whether they're needed for quests.");
		else
			GFWUtils.Print("Usage: "..GFWUtils.Hilite("/feedomatic savequest on").." | "..GFWUtils.Hilite("off"));
		end
		return;
	end
	
	-- Set avoiding quest-objective food
	if ( cmd == "savebonus" ) then
		if (option == "on") then
			FOM_Config.AvoidBonusFood = true;
			FOM_Config.AvoidUsefulFood = true;
			GFWUtils.Print("Feed-O-Matic will avoid foods that have an additional bonus effect when eaten by a player.");
		elseif (option == "off") then
			FOM_Config.AvoidBonusFood = true;
			if not (FOM_Config.SaveForCookingLevel >= 0 and FOM_Config.SaveForCookingLevel <= 3 and not FOM_Config.AvoidQuestFood) then
				FOM_Config.AvoidUsefulFood = false;
			end
			GFWUtils.Print("Feed-O-Matic will choose foods without regard to whether they have bonus effects.");
		else
			GFWUtils.Print("Usage: "..GFWUtils.Hilite("/feedomatic savebonus on").." | "..GFWUtils.Hilite("off"));
		end
		return;
	end
	
	-- Set avoiding quest-objective food
	if ( cmd == "fallback" ) then
		if (option == "on") then
			FOM_Config.Fallback = true;
			GFWUtils.Print("Feed-O-Matic will fall back to food it would otherwise avoid if no other food is available.");
		elseif (option == "off") then
			FOM_Config.Fallback = false;
			GFWUtils.Print("Feed-O-Matic will not feed your pet if the only foods available are foods you'd prefer to avoid feeding.");
		else
			GFWUtils.Print("Usage: "..GFWUtils.Hilite("/feedomatic fallback on").." | "..GFWUtils.Hilite("off"));
		end
		return;
	end
	
	-- Set quality sorting direction
	if ( cmd == "quality" ) then
		if (option == "high") then
			FOM_Config.PreferHigherQuality = true;
			GFWUtils.Print("Feed-O-Matic will prefer to use higher quality foods first.");
		elseif (option == "low") then
			FOM_Config.PreferHigherQuality = false;
			GFWUtils.Print("Feed-O-Matic will prefer to use lower quality foods first.");
		else
			GFWUtils.Print("Usage: "..GFWUtils.Hilite("/feedomatic quality high").." | "..GFWUtils.Hilite("low"));
		end
		return;
	end

	-- Set inventory management threshold
	if ( cmd == "keepopen" ) then
		if (option == "off" or option == "none") then
			newNum = 0;
		elseif (option == "max") then
			newNum = MAX_KEEPOPEN_SLOTS;
		else
			newNum = tonumber(option);
		end
		if (newNum == nil) then
			GFWUtils.Print("Usage: "..GFWUtils.Hilite("/feedomatic keepopen <number>"));
			return;
		end
		FOM_Config.KeepOpenSlots = newNum;
		GFWUtils.Print("Feed-O-Matic will try to keep at least "..GFWUtils.Hilite(FOM_Config.KeepOpenSlots).." spaces open in your inventory when looking for food.");
		return;
	end
	
	-- Feed Pet
	local _, _, cmd, foodString = string.find(msg, "(%w+) *(.*)");
	if ( cmd == "feed" ) then
		if (foodString == "") then
			FOM_Feed(nil); -- automatically find a food and feed it
		else
			local inputFoods = { };
			for itemLink in string.gfind(foodString, "%[[%w%s:()\"'-]+%]") do
				local _, _, foodName = string.find(itemLink, "^%[([%w%s:()\"'-]+)%]$"); 
				table.insert(inputFoods, foodName);
			end
			if (table.getn(inputFoods) == 0) then
				table.insert(inputFoods, foodString); -- if no item links, treat whole input line as one food's name
			end
			
			for _, food in inputFoods do
				FOM_Feed(food);
			end
		end
		return;
	end
	
	local _, _, cmd, diet, foodString = string.find(msg, "(%w+) (%w+) *(.*)");
	if ( cmd ~= nil ) then
	
		diet = string.lower(diet); -- let's be case insensitive
		if ( FOM_Foods[diet] == nil and diet ~= "all") then
			local usageString = "Usage: "..GFWUtils.Hilite("/feedomatic add meat").." | "..GFWUtils.Hilite("fish").." | "..GFWUtils.Hilite("bread").." | "..GFWUtils.Hilite("cheese").." | "..GFWUtils.Hilite("fruit").." | "..GFWUtils.Hilite("fungus").." | "..GFWUtils.Hilite("bonus")
			if (cmd == "show" or cmd == "list") then
				usageString = usageString.." <food>.";
			end
			GFWUtils.Print(usageString);
			return;
		end

		if (cmd == "show" or cmd == "list") then
			if ( diet == "all" ) then
				diets = {"meat", "fish", "bread", "cheese", "fruit", "fungus", "bonus"};
			else
				diets = {diet};
			end
			
			for _, aDiet in diets do
				local capDiet = string.upper(string.sub(aDiet, 1, 1)) .. string.sub(aDiet, 2); -- print a nicely capitalized version
				GFWUtils.Print("Feed-O-Matic "..GFWUtils.Hilite(capDiet).." List:");
				local dietFoods = FOM_Foods[aDiet];
				if (FOM_AddedFoods) then
					dietFoods = GFWTable.Merge(dietFoods, FOM_AddedFoods[aDiet]);
				end
				if (FOM_RemovedFoods) then
					dietFoods = GFWTable.Subtract(dietFoods, FOM_RemovedFoods[aDiet]);
				end
				table.sort(dietFoods);
				for _, food in dietFoods do
					GFWUtils.Print(GFWUtils.Hilite(" - ")..food);
				end
			end
			return;
		else
		
			local inputFoods = { };
			for itemLink in string.gfind(foodString, "%[[%w%s:()\"'-]+%]") do
				local _, _, foodName = string.find(itemLink, "^%[([%w%s:()\"'-]+)%]$"); 
				table.insert(inputFoods, foodName);
			end
			if (table.getn(inputFoods) == 0) then
				table.insert(inputFoods, foodString); -- if no item links, treat whole input line as one food's name
			end

			local capDiet = string.upper(string.sub(diet, 1, 1)) .. string.sub(diet, 2); -- print a nicely capitalized version
			if ( cmd == "add" ) then
				for _, food in inputFoods do
					if ( FOM_AddFood(diet, food) ) then
						GFWUtils.Print("Added "..GFWUtils.Hilite(food).." to "..GFWUtils.Hilite(capDiet).." list.");
					else
						GFWUtils.Print(GFWUtils.Hilite(food).." already in "..GFWUtils.Hilite(capDiet).." list.");
					end
				end
				if (FOM_Config.AvoidQuestFood) then
					FOM_ScanQuests(); -- in case any of the newly added foods are quest objectives
				end
				return;
			elseif (cmd == "remove" ) then
				for _, food in inputFoods do
					if ( FOM_RemoveFood(diet, food) ) then
						GFWUtils.Print("Removed "..GFWUtils.Hilite(food).." from "..GFWUtils.Hilite(capDiet).." list.");
					else
						GFWUtils.Print("Could not find "..GFWUtils.Hilite(food).." in "..GFWUtils.Hilite(capDiet).." list.");
					end
				end
				return;
			end
		end
	end
	
	-- if we got down to here, we got bad input
	FOM_ChatCommandHandler("help");
end

-- Add a food to a list
function FOM_AddFood(diet, food)

	if (FOM_Foods[diet] == nil) then
		GFWUtils.DebugLog("FOM_Foods[diet] == nil");
	end
	if (FOM_AddedFoods == nil or FOM_AddedFoods[diet] == nil) then
		GFWUtils.DebugLog("FOM_AddedFoods == nil or FOM_AddedFoods[diet] == nil");
	end
	if (FOM_RemovedFoods == nil or FOM_RemovedFoods[diet] == nil) then
		GFWUtils.DebugLog("FOM_RemovedFoods == nil or FOM_RemovedFoods[diet] == nil");
	end
	if ( GFWTable.IndexOf(FOM_Foods[diet], food) == 0 ) then
		if (FOM_AddedFoods == nil) then
			FOM_AddedFoods = {};
		end
		if (FOM_AddedFoods[diet] == nil) then
			FOM_AddedFoods[diet] = {};
		end
		if ( GFWTable.IndexOf(FOM_AddedFoods[diet], food) == 0 ) then
			table.insert( FOM_AddedFoods[diet], food );
			table.sort( FOM_AddedFoods[diet] );
			if (FOM_RemovedFoods and FOM_RemovedFoods[diet] and GFWTable.IndexOf(FOM_RemovedFoods[diet], food) ~= 0) then
				table.remove( FOM_RemovedFoods[diet], GFWTable.IndexOf(FOM_RemovedFoods[diet], food) );
				table.sort( FOM_RemovedFoods[diet] );
			end
			return true;
		else
			return false;
		end
	else
		return false;
	end

end

-- Remove a food from a list
function FOM_RemoveFood(diet, food)
	
	if (FOM_Foods[diet] == nil) then
		GFWUtils.DebugLog("FOM_Foods[diet] == nil");
	end
	if (FOM_AddedFoods == nil or FOM_AddedFoods[diet] == nil) then
		GFWUtils.DebugLog("FOM_AddedFoods == nil or FOM_AddedFoods[diet] == nil");
	end
	if (FOM_RemovedFoods == nil or FOM_RemovedFoods[diet] == nil) then
		GFWUtils.DebugLog("FOM_RemovedFoods == nil or FOM_RemovedFoods[diet] == nil");
	end
	if ( GFWTable.IndexOf(FOM_Foods[diet], food) ~= 0 ) then
		if (FOM_RemovedFoods == nil) then
			FOM_RemovedFoods = {};
		end
		if (FOM_RemovedFoods[diet] == nil) then
			FOM_RemovedFoods[diet] = {};
		end
		if ( GFWTable.IndexOf(FOM_RemovedFoods[diet], food) == 0 ) then
			table.insert( FOM_RemovedFoods[diet], food );
			table.sort( FOM_RemovedFoods[diet] );
			if (FOM_AddedFoods and FOM_AddedFoods[diet] and GFWTable.IndexOf(FOM_AddedFoods[diet], food) ~= 0) then
				table.remove( FOM_AddedFoods[diet], GFWTable.IndexOf(FOM_AddedFoods[diet], food) );
				table.sort( FOM_AddedFoods[diet] );
			end
			return true;
		else
			return false;
		end
	else
		return false;
	end

end

-- Check Happiness
function FOM_CheckHappiness()

	-- Check for pet
	if not ( UnitExists("pet") ) then
		return;
	end
	
	-- Check if AFK
	if ( FOM_State.IsAFK ) then
		return;
	end
	
	if ( (GetBattlefieldStatus()) == "active" ) then
		-- don't auto-feed while in a battleground; the frequent spirit rez gives mucho happiness (slightly delayed)
		return;
	end
	
	-- Check if in Combat
	if ( FOM_State.InCombat ) or ( PlayerFrame.inCombat ) then
		return;
	end
	
	-- Check if Has Feed Effect
	if ( FOM_HasFeedEffect() ) then
		return;
	end

	-- Get Pet Info
	local pet = UnitName("pet");
	local happiness, damage, loyalty = GetPetHappiness();
	
	local level;
	if ( FOM_Config.Level == "content" ) then
		level = 2;
	elseif ( FOM_Config.Level == "happy" ) then
		level = 3;
	elseif ( FOM_Config.Level == "debug" ) then
		level = 4;
	end
	
	-- Check No Happiness
	if ( happiness == 0 ) or ( happiness == nil ) then
		return;
	end
	
	-- Check if Need Feeding
	if ( happiness < level ) then
		if (FOM_FeedLock.Locked == false and FOM_CanFeed() ) then
			FOM_FeedLock.Locked = true;
			FOM_State.ShouldFeed = false;
			GFWUtils.DebugLog("Feeding on event");
			FOM_Feed(nil);
		else
			FOM_State.ShouldFeed = true;
		end
	end
	
end

-- Feed Pet
function FOM_Feed(aFood)
	
	FOM_State.ShouldFeed = false;
	
	-- Make sure we have a feedable pet
	if not (UnitExists("pet")) then 
		GFWUtils.Note("You don't have a pet to feed."); 
		return;
	end
	if (UnitIsDead("pet")) then
		GFWUtils.Note("Your pet is dead."); 
		return;
	end
	if (GetPetFoodTypes() == nil) then
		GFWUtils.Note("You don't have a feedable pet."); 
		return;
	end
	
	-- Assign Variable
	local pet = UnitName("pet");
	
	FOM_CheckSetup();
	
	-- Look for Food
	local foodBag, foodItem;
	if (aFood ~= nil) then
		-- if told to feed a specific food, do so
		foodBag, foodItem = FOM_FindSpecificFood(aFood);
		if ( foodBag == nil) then
			-- No Food Could be Found
			GFWUtils.Print(pet.." could not find "..aFood.." in your pack.");
			return;
		end
	else
		foodBag, foodItem = FOM_NewFindFood();
	end
	
	if ( foodBag == nil) then
		-- No Food Could be Found
		GFWUtils.Print(pet.." could not find any food in your pack.");
		FOM_State.OutOfFoodTime = GetTime();
		return;
	end
		
	LastFood = FOM_GetItemName(foodBag, foodItem);
	
	if (FOM_Config.Debug) then
		-- don't actually feed anything, just show what we would choose
		GFWUtils.DebugLog("Picked "..LastFood.." (bag "..foodBag..", slot "..foodItem..") for feeding.");
		return;
	end
	
	-- Actually feed the item to the pet
	PickupContainerItem(foodBag, foodItem);
	if ( CursorHasItem() ) then
		DropItemOnUnit("pet");
	end
	if ( CursorHasItem() ) then
		PickupContainerItem(foodBag, foodItem);
	else
		-- Alert
		if ( FOM_Config.Alert == "chat") then
			GFWUtils.Print(pet.." eats a "..GFWUtils.Hilite(LastFood).." from your pack.");
		elseif ( FOM_Config.Alert == "emote") then
			SendChatMessage("feeds ".. pet .." a ".. LastFood ..". ".. FOM_RandomEmote(), "EMOTE");
		end
	end
		
end

function FOM_RandomEmote()
	
	local randomEmotes = {};
	if (UnitSex("pet") == 0) then
		randomEmotes = GFWTable.Merge(randomEmotes, FOM_Emotes["male"]);
	elseif (UnitSex("pet") == 1) then
		randomEmotes = GFWTable.Merge(randomEmotes, FOM_Emotes["female"]);
	end
	
	randomEmotes = GFWTable.Merge(randomEmotes, FOM_Emotes[UnitCreatureFamily("pet")]);
	randomEmotes = GFWTable.Merge(randomEmotes, FOM_Emotes[LastFood]);
	randomEmotes = GFWTable.Merge(randomEmotes, FOM_Emotes["any"]);
	
	return randomEmotes[math.random(table.getn(randomEmotes))];

end

function FOM_FindSpecificFood(foodName)
	for bagNum = 0, 4 do
		local bagName = GetBagName(bagNum);
		if ((bagName ~= nil) and not (strfind(bagName, "Ammo") or strfind(bagName, "Quiver")) ) then
		-- skip bags that can't contain food
		
			local bagSize = GetContainerNumSlots(bagNum);
			for itemNum = 1, bagSize do
		
				itemName = FOM_GetItemName(bagNum, itemNum);
				if ( itemName == foodName and FOM_IsInDiet(itemName) ) then
					return bagNum, itemNum;
				end
			
			end
		end
	end
	return nil;
end

function FOM_FindSmallestFoodStack(desiredQuality, avoidUseful)
	local smallestStackBag = -1;
	local smallestStackItem = -1;
	local smallestStackSize = 201; -- arbitrarily large, in case there might ever be food that stacks as dense as ammo does.
	
	if (desiredQuality == nil) then
		GFWUtils.DebugLog("Looking for food of unknown quality...");
	else
		GFWUtils.DebugLog("Looking for food of quality "..desiredQuality.." or better...");
	end
	if (avoidUseful) then
		GFWUtils.DebugLog("Avoiding 'useful' food...");
	else
		GFWUtils.DebugLog("Who cares if it's got another use, I'm hungry!");
	end

	FOM_Quantity = { };
	for bagNum = 0, 4 do
		local bagName = GetBagName(bagNum);
		if ((bagName ~= nil) and not (strfind(bagName, "Ammo") or strfind(bagName, "Quiver")) ) then
		-- skip bags that can't contain food
		
			local bagSize = GetContainerNumSlots(bagNum);
			for itemNum = 1, bagSize do
		
				local itemName = FOM_GetItemName(bagNum, itemNum);
				if ( FOM_IsInDiet(itemName) ) then
					local _, itemCount = GetContainerItemInfo(bagNum, itemNum);
					if (avoidUseful and FOM_IsUsefulFood(itemName, itemCount)) then
						-- skip to next
					else
						foodQuality = FOM_FoodQuality[RealmPlayer][LastPetName][itemName];
						GFWUtils.DebugLog("Quality for "..itemName..": "..(foodQuality or "(nil)"));
						if (foodQuality == nil) then
							if (desiredQuality == nil) then
								if ( itemCount < smallestStackSize ) then
									smallestStackSize = itemCount;
									smallestStackBag = bagNum;
									smallestStackItem = itemNum;
								end
							end
						elseif (desiredQuality ~= nil and foodQuality >= desiredQuality) then
							GFWUtils.DebugLog(itemCount.. "x "..itemName.." meets desired quality "..desiredQuality..". Smallest stack:"..smallestStackSize );
							if ( itemCount < smallestStackSize ) then
								smallestStackSize = itemCount;
								smallestStackBag = bagNum;
								smallestStackItem = itemNum;
							end
						end
					end
				end
			
			end
		end
	end
	
	return smallestStackBag, smallestStackItem;
end

function FOM_IsTemporaryFood(foodName)
	if (foodName ~= nil and string.find(foodName, "^Conjured")) then
		return 1;
	else
		return 0;
	end
end

function FOM_FlatFoodList()
	local foodList = {};
	FOM_Quantity = { };
	for bagNum = 0, 4 do
		local bagName = GetBagName(bagNum);
		if ((bagName ~= nil) and not (strfind(bagName, "Ammo") or strfind(bagName, "Quiver")) ) then
		-- skip bags that can't contain food
			for itemNum = 1, GetContainerNumSlots(bagNum) do
				local itemName = FOM_GetItemName(bagNum, itemNum);
				local _, itemCount = GetContainerItemInfo(bagNum, itemNum);
				if ( FOM_IsInDiet(itemName) ) then
					local isUseful = FOM_IsUsefulFood(itemName, itemCount);
					foodQuality = (FOM_FoodQuality[RealmPlayer][LastPetName][itemName] or MAX_QUALITY);
					table.insert(foodList, {bag=bagNum, slot=itemNum, name=itemName, count=itemCount, quality=foodQuality, useful=isUseful, temp=FOM_IsTemporaryFood(itemName)});
				end
			end
		end
	end
	return foodList;
end

function FOM_NewFindFood()
	FlatFoodList = FOM_FlatFoodList();
	
	table.sort(FlatFoodList, FOM_SortCount); -- small stacks first
	if (FOM_NumOpenBagSlots() > FOM_Config.KeepOpenSlots) then
		if (FOM_Config.PreferHigherQuality) then
			table.sort(FlatFoodList, FOM_SortQualityDescending); -- higher quality first
		else
			table.sort(FlatFoodList, FOM_SortQualityAscending); -- lower quality first
		end
	end
	table.sort(FlatFoodList, FOM_SortTemporary); -- temporary foods first
	if (FOM_Config.AvoidUsefulFood) then
		table.sort(FlatFoodList, FOM_SortUseful); -- non-useful foods first
	end
		
	for _, foodInfo in FlatFoodList do
		if (foodInfo.useful and FOM_Config.AvoidUsefulFood and not FOM_Config.Fallback) then
			GFWUtils.DebugLog("Skipping "..quantity.."x "..foodName.."; no falling back to avoided foods.");
		else
			return foodInfo.bag, foodInfo.slot;
		end
	end
	
	return nil;
end

function FOM_SortTemporary(a, b)
	return a.temp > b.temp;
end

function FOM_SortCount(a, b)
	return a.count < b.count;
end

function FOM_SortUseful(a, b)
	if (a.useful) then
		aUseful = 1;
	else
		aUseful = 0;
	end
	if (b.useful) then
		bUseful = 1;
	else
		bUseful = 0;
	end
	return aUseful < bUseful;
end

function FOM_SortQualityDescending(a, b)
	return a.quality > b.quality;
end

function FOM_SortQualityAscending(a, b)
	return a.quality < b.quality;
end

function FOM_IsUsefulFood(foodName, quantity)
	if (FOM_Cooking and FOM_Cooking[RealmPlayer] and FOM_Cooking[RealmPlayer][foodName]) then
		if (FOM_Cooking[RealmPlayer][foodName] >= FOM_Config.SaveForCookingLevel) then
			GFWUtils.DebugLog("Skipping "..quantity.."x "..foodName.."; is good for cooking.");
			return true;
		end
	end
	if (FOM_Config.AvoidQuestFood) then
		FOM_ScanQuests();
		if (FOM_QuestFood ~= nil and FOM_QuestFood[RealmPlayer] ~= nil and FOM_QuestFood[RealmPlayer][foodName]) then
			if (FOM_Quantity[foodName] == nil) then
				FOM_Quantity[foodName] = quantity;
			else
				FOM_Quantity[foodName] = FOM_Quantity[foodName] + quantity;
			end
			if (FOM_Quantity[foodName] > FOM_QuestFood[RealmPlayer][foodName]) then
				GFWUtils.DebugLog("Not skipping "..quantity.."x "..foodName.."; is needed for quest, but we have more than enough.");
				return false;
			else
				GFWUtils.DebugLog("Skipping "..quantity.."x "..foodName.."; is needed for quest.");
				return true;
			end
		end
	end
	if (FOM_Config.AvoidBonusFood and FOM_IsInDiet(foodName, "bonus")) then
		GFWUtils.DebugLog("Skipping "..quantity.."x "..foodName.."; has bonus effect when eaten by player.");
		return true;
	end
	--GFWUtils.DebugLog("Not skipping "..quantity.."x "..foodName.."; doesn't have other uses.");
	return false;
end

function FOM_ReportUnusedFoods()

	FOM_CheckSetup();
	
	local unusedFoods = { };
	for bagNum = 0, 4 do
		local bagName = GetBagName(bagNum);
		if ((bagName ~= nil) and not (strfind(bagName, "Ammo") or strfind(bagName, "Quiver")) ) then
		-- skip bags that can't contain food
		
			local bagSize = GetContainerNumSlots(bagNum);
			for itemNum = 1, bagSize do
				local itemName = FOM_GetItemName(bagNum, itemNum);
				local _, itemCount = GetContainerItemInfo(bagNum, itemNum);
				if (not FOM_IsUsefulFood(itemName, itemCount)) then
					if (FOM_FoodQuality[RealmPlayer][LastPetName][itemName] == 0) then
						table.insert(unusedFoods, itemName);
					end
				end
			end
		end
	end
	
	if (table.getn(unusedFoods) > 0) then
		GFWUtils.Print("Unused foods: "..table.concat(unusedFoods, ", "));
	end
end

function FOM_NumOpenBagSlots()
	local openSlots = 0;
	for bagNum = 0, 4 do
		local bagName = GetBagName(bagNum);
		if ((bagName ~= nil) and not (strfind(bagName, "Ammo") or strfind(bagName, "Quiver")) ) then
		-- skip bags that can't contain food

			local bagSize = GetContainerNumSlots(bagNum);
			for itemNum = 1, bagSize do
				if (GetContainerItemInfo(bagNum, itemNum) == nil) then
					openSlots = openSlots + 1;
				end
			end
		end
	end
	return openSlots;
end

-- Check Feed Effect
function FOM_HasFeedEffect()

	local i = 1;
	local buff;
	buff = UnitBuff("pet", i);
	while buff do
		if ( string.find(buff, "Ability_Hunter_BeastTraining") ) then
			return true;
		end
		i = i + 1;
		buff = UnitBuff("pet", i);
	end
	return false;

end

-- Check Food Types
function FOM_IsInDiet(food, dietList)

	if ( dietList == nil ) then
		dietList = {GetPetFoodTypes()};
	end
	if ( dietList == nil ) then
		return false;
	end
	if (type(dietList) ~= "table") then
		dietList = {dietList};
	end
	for _, diet in dietList do 
		diet = string.lower(diet); -- let's be case insensitive
		if (FOM_Foods[diet] == nil) then
			GFWUtils.DebugLog("FOM_Foods[diet] == nil");
		end
		if (FOM_RemovedFoods ~= nil and FOM_RemovedFoods[diet] ~= nil and GFWTable.IndexOf(FOM_RemovedFoods[diet], food) ~= 0) then
			return false;
		end
		if (FOM_AddedFoods ~= nil and FOM_AddedFoods[diet] ~= nil and GFWTable.IndexOf(FOM_AddedFoods[diet], food) ~= 0) then
			return true;
		end
		if (GFWTable.IndexOf(FOM_Foods[diet], food) ~= 0) then
			return true;
		end
	end
	
	return false;

end

function FOM_IsKnownFood(food)
	return FOM_IsInDiet(food, {"meat", "fish", "bread", "cheese", "fungus", "fruit"});
end

-- Get Item Name
function FOM_GetItemName(bag, slot)

	local itemLink = GetContainerItemLink(bag, slot);
	if (itemLink) then
		local _, _, name = string.find(itemLink, "%[(.+)%]"); 
		return name;
	else
		return "";
	end
end

-- Convert info saved by older versions / Rauen's PetFeed
function FOM_LoadOldData()
	if (PetFeed_Foods_Meat ~= nil) then
		for food in PetFeed_Foods_Meat do
			FOM_AddFood("meat", food);
		end
		PetFeed_Foods_Meat = nil;
	end
	if (PetFeed_Foods_Fish ~= nil) then
		for food in PetFeed_Foods_Fish do
			FOM_AddFood("fish", food);
		end
		PetFeed_Foods_Fish = nil;
	end
	if (PetFeed_Foods_Bread ~= nil) then
		for food in PetFeed_Foods_Bread do
			FOM_AddFood("bread", food);
		end
		PetFeed_Foods_Bread = nil;
	end
	if (PetFeed_Foods_Cheese ~= nil) then
		for food in PetFeed_Foods_Cheese do
			FOM_AddFood("cheese", food);
		end
		PetFeed_Foods_Cheese = nil;
	end
	if (PetFeed_Foods_Fruit ~= nil) then
		for food in PetFeed_Foods_Fruit do
			FOM_AddFood("fruit", food);
		end
		PetFeed_Foods_Fruit = nil;
	end
	if (PetFeed_Foods_Fungus ~= nil) then
		for food in PetFeed_Foods_Fungus do
			FOM_AddFood("fungus", food);
		end
		PetFeed_Foods_Fungus = nil;
	end
	return;
end

function FOM_PetRename(newName)

	FOM_CheckSetup();

	-- move our saved food quality data to be indexed under the new name
	FOM_FoodQuality[RealmPlayer][newName] = FOM_FoodQuality[RealmPlayer][LastPetName];
	FOM_FoodQuality[RealmPlayer][LastPetName] = nil;

	FOM_Original_PetRename(newName);
	
end

function FOM_PetAbandon()
	
	FOM_CheckSetup();

	-- delete saved food-quality data for this pet so we don't bloat SavedVariables
	FOM_FoodQuality[RealmPlayer][LastPetName] = nil;

	FOM_Original_PetAbandon();
	
end
